
## CS294: Practical Machine Learning (Fall 09, modified from Spr 2008)
## PSet3, Hidden Markov Models (script with sample code)

# activate hmm library, and load auxiliary functions
library('hmm.discnp');
source('hw4-hmmFuncs.R');

# load training & test datasets
trainText <- readText('aliceTrain.txt');
testText  <- readText('aliceTest.txt');

##
## Part I:  EM Training & Model Order Selection
##

# train models of varying orders
Nstate <- c(1,3,5,10,15,20,30);
# (feel free to try larger state dimensions, but note that the
# hmm.discnp package is VERY inefficient in its memory allocation) 

hmmAll <- list();
trainLogLike <- c();
for (kk in 1:length(Nstate)) {
  cat(paste("\n*** Training with ", Nstate[kk], " states: ***\n", sep = ""));
  hmmAll[[kk]] <- hmm(y=trainText, 
    K=Nstate[kk], tolerance=1e-4, itmax=500, verbose=TRUE);
  trainLogLike[kk] <- hmmAll[[kk]]$log.like;
}

# compute AIC/BIC scores & test likelihoods for all models 
trainAIC <- c();
trainBIC <- c();
testLogLike <- c();
for (kk in 1:length(Nstate)) {
  #TODO: Compute the number of free parameters, and the BIC score
  #numParam <- ???
  #trainBIC[kk] <- ???
  trainAIC[kk] <- trainLogLike[kk] - numParam;

  testLogLike[kk] <- loglike.hmm(y=testText, yval=levels(testText),
    tpm=hmmAll[[kk]]$tpm, Rho=hmmAll[[kk]]$Rho);
}
#TODO: Plot the AIC scores, BIC scores, and training likelihoods
#TODO: Plot test likelihoods

# select best scoring models 
bestBIC <- which.max(trainBIC);
bestAIC <- which.max(trainAIC);

# simulate random text strings from trained models 
sampleText(500, hmmAll[[1]], levels(trainText))
sampleText(500, hmmAll[[bestBIC]], levels(trainText))
sampleText(500, hmmAll[[bestAIC]], levels(trainText))

##
## Part II:  Filling in Deleted Letters
##

# randomly remove subset of letters from test sentence
eraseRate <- 0.2;
eraseProb <- rbind(diag(1-eraseRate, nrow=27), eraseRate);
noisyText <- perturbText(testText, eraseProb);
numErased <- sum(noisyText == "*");

# test each of the 3 models: no states, BIC, AIC
hmmTest <- hmmAll[c(1,bestBIC,bestAIC)];
denoisedText <- list();
numCorrected <- c();

for (kk in 1:3) {
  # determine likelihood of observations given each hidden state
  obsProb <- eraseProb %*% hmmTest[[kk]]$Rho;

  # compute posterior probability of hidden state sequence
  stateProb <- margprob.hmm(noisyText, levels(noisyText), 
    hmmTest[[kk]]$tpm, obsProb);

  # find most likely value of each erased letter
  denoisedText[[kk]] <- noisyText;
  for (cc in 1:length(denoisedText[[kk]])) {
    if (denoisedText[[kk]][cc]=="*") {
      #TODO: Implement optimal estimator for missing letter
      #denoisedText[[kk]][cc] <- ???
    }
  }

  # determine number of corrected erasures
  numCorrected[kk] <- numErased - sum(denoisedText[[kk]] != testText);
}

# display denoised text for each model
noisyText[1:500]
denoisedText[[1]][1:500]
denoisedText[[2]][1:500]
denoisedText[[3]][1:500]

